

#define MK2CODENET_VERSION "0.1"

#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>

#ifdef _WIN32
#include <windows.h>
#else
#include <unistd.h>
#endif

#include "RS-232/rs232.h" 
#include "RS-232/rs232.c" 

#define MK2BAUDRATE 19200
int cport_nr = 0;

int verbose = 0;
#define C64_RAM_SIZE    0x200000 /* includes RAM used for ROMs etc */
unsigned char *buffer = NULL;

#define LOGLVL_MSG 0
#define LOGLVL_VER 1
#define LOGLVL_ERR 2 
#define LOGLVL_DBG 3

#define LOGERR(...)       logfunc (LOGLVL_ERR, __VA_ARGS__ )
#define LOGVER(...)       logfunc (LOGLVL_VER, __VA_ARGS__ )
#define LOGMSG(...)       logfunc (LOGLVL_MSG, __VA_ARGS__ )
#define LOGDBG(...)       logfunc (LOGLVL_DBG, __VA_ARGS__ )

int cleanup(int n);

#define DELAY_RWMEM       10   /* delay after each block/line */
#define DELAY_PROMPT    1000   /* delay when waiting for a prompt */
#define DELAY_COMMAND   2000   /* delay before and after sending a command */
#define DELAY_POLL      1000   /* delay when polling memory */

#define WRITE_BLOCK_LEN 64

int logfunc (int lvl, const char * format, ...)
{
    int printed = 0;
    va_list ap;
    va_start (ap, format);

    switch (lvl) {
        case LOGLVL_ERR:
            printed = vfprintf (stderr, format, ap);
            break;
        case LOGLVL_VER:
            if (verbose >= 1) printed = vfprintf (stdout, format, ap);
            break;
        case LOGLVL_MSG:
            printed = vfprintf (stdout, format, ap);
            break;
        case LOGLVL_DBG:
            if (verbose >= 2) printed = vfprintf (stdout, format, ap);
            break;
    }

    va_end (ap);
    fflush(stdout);
    return printed;
}

void mk2_usleep(unsigned int us)
{
#ifdef _WIN32
    Sleep(us / 1000);
#else
    usleep(us);
#endif
}

int mk2_putbyte(unsigned char c)
{
    int n;
//    printf(">mk2_putbyte\n");
    n = RS232_SendByte(cport_nr, c);
//    printf("<mk2_putbyte (%d,%c)\n",n,c);
    return n;
}

int mk2_getbyte(void)
{
    unsigned char c;
    int n;
    while ((n = RS232_PollComport(cport_nr, &c, 1)) < 1) { /* wait */ }
    LOGDBG("got %02x (%d)\n",c,n);
    return c;
}

int mk2_putbyte_getecho(unsigned char c)
{
    unsigned char e;
    int n;
    n = mk2_putbyte(c);
    e = mk2_getbyte();
    if (e != c) {
        LOGERR("error: got unexpected echoed value (got %02x, expected %02x)\n", e, c);
        return -1;
    }
    return n;
}

int mk2_waitprompt(void)
{
    unsigned char c;
    int n;
    unsigned long timeout = 10000000 / DELAY_PROMPT;

    LOGDBG(">waitprompt\n");

    while(timeout--) {
        n = RS232_PollComport(cport_nr, &c, 1);
        if(n > 0) {
            LOGVER("%c", c);
            if (c == '?') {
                RS232_flushRXTX(cport_nr);
                LOGDBG("<waitprompt\n");
                return 1;
            }
        } else {
//            mk2_putbyte(' ');
            mk2_usleep(DELAY_PROMPT);
        }
    }

    LOGERR("error: timeout waiting for data.\n");
    exit(cleanup(EXIT_FAILURE));
    return n;
}

int mk2_sendcommand(unsigned char c)
{
    int n;
    mk2_usleep(DELAY_COMMAND);
    LOGDBG("cmd: %c\n",c);
    n = mk2_putbyte(c);
    mk2_usleep(DELAY_COMMAND);
    return n;
}

int mk2_mainmenu(void)
{
    int i, n;
    LOGDBG(">mk2_mainmenu\n");
    
    RS232_flushRXTX(cport_nr);

    n = mk2_sendcommand(' ');
    n = mk2_sendcommand('x');
    for(i = 0; i < 10; i++) {
        mk2_waitprompt();
        n = mk2_sendcommand('x');
    }
    LOGDBG("<mk2_mainmenu\n");
    return n;
}

int mk2_writememory(unsigned char *buf, unsigned int len, unsigned int addr)
{
    unsigned char b[8];
    int i;
    mk2_mainmenu();
    mk2_waitprompt();
    mk2_sendcommand('d');
    mk2_waitprompt();
    mk2_sendcommand('w');
    mk2_waitprompt();
    sprintf((char*)b, "%05X", addr);
    LOGDBG("addr: %s\n",b);
    mk2_putbyte_getecho(b[0]);
    mk2_putbyte_getecho(b[1]);
    mk2_putbyte_getecho(b[2]);
    mk2_putbyte_getecho(b[3]);
    mk2_putbyte_getecho(b[4]);
    mk2_putbyte(0x0d);
    mk2_waitprompt();
    for (i = 0;i < len; i++) {
        sprintf((char*)b, "%02X", buf[i]);
        LOGDBG("send %s (%d)\n",b,i);
        mk2_putbyte_getecho(b[0]);
        mk2_putbyte_getecho(b[1]);
        if ((i % WRITE_BLOCK_LEN) == (WRITE_BLOCK_LEN - 1)) {
            LOGDBG("next (%d)\n",i);
            mk2_putbyte(0x0d);
            mk2_waitprompt();
            mk2_usleep(DELAY_RWMEM);
        }
    }
    mk2_putbyte(0x0d);
    mk2_waitprompt();
    mk2_sendcommand('x');
    mk2_waitprompt();
    mk2_sendcommand('c');
    
    return 0;
}

int mk2_readmemory(unsigned char *buf, unsigned int len, unsigned int addr)
{
    unsigned char b[8];
    int i;
    mk2_mainmenu();
    mk2_waitprompt();
    mk2_sendcommand('d');
    mk2_waitprompt();
    mk2_sendcommand('r');

    mk2_waitprompt();
    sprintf((char*)b, "%05X", addr);
    LOGDBG("addr: %s\n",b);
    mk2_putbyte_getecho(b[0]);
    mk2_putbyte_getecho(b[1]);
    mk2_putbyte_getecho(b[2]);
    mk2_putbyte_getecho(b[3]);
    mk2_putbyte_getecho(b[4]);
    mk2_putbyte(0x0d);

    mk2_waitprompt();
    sprintf((char*)b, "%05X", addr + len - 1);
    LOGDBG("addr: %s\n",b);
    mk2_putbyte_getecho(b[0]);
    mk2_putbyte_getecho(b[1]);
    mk2_putbyte_getecho(b[2]);
    mk2_putbyte_getecho(b[3]);
    mk2_putbyte_getecho(b[4]);
    mk2_putbyte(0x0d);

    b[2] = 0;
    for (i = 0;i < len; i++) {
        /* skip whitespace */
        while ((b[0] = mk2_getbyte()) <= 0x20) {}
        b[1] = mk2_getbyte();
        LOGDBG("gotbyte %s (%d)\n",b,i);
        buf[i] = strtoul((char*)b, NULL, 16);
        mk2_usleep(DELAY_RWMEM);
    }
    mk2_waitprompt();
    mk2_sendcommand('c');
    
    return 0;
}

void execute_reset(void)
{
    LOGDBG(">execute_reset\n");
    mk2_mainmenu();
    mk2_waitprompt();
    mk2_sendcommand('r');
    LOGDBG("<execute_reset\n");
}

void execute_resetwait(void)
{
    unsigned char buf[6];
    unsigned int readyloc = 1024 + (5 * 40);

    /* mk2_mainmenu(); */ /* unneeded, mk2_writememory will do it */

    /* overwrite the memory where we expect "ready" after reset */
    buf[0] = 32;
    buf[1] = 32;
    buf[2] = 32;
    buf[3] = 32;
    buf[4] = 32;
    buf[5] = 32;
    if (mk2_writememory(buf, 6, readyloc) < 0) {
        LOGERR("error writing to C64 memory.\n");
        exit(cleanup(EXIT_FAILURE));
    }
    execute_reset();
    printf("waiting for reset to finish..."); fflush(stdout);
    while ((buf[0] != 18) && (buf[1] != 5) && (buf[2] != 1) &&
           (buf[3] != 4) && (buf[4] != 25) && (buf[5] != 46)){
        if (mk2_readmemory(buf, 6, readyloc) < 0) {
            LOGERR("error reading C64 memory.\n");
            exit(cleanup(EXIT_FAILURE));
        }
        mk2_usleep(DELAY_POLL);
    }
    printf("done.\n");
    
}

void setup_pointers(int addr, int len)
{
    unsigned char buf[10];
    buf[0] = (addr + len) & 0xff;
    buf[1] = ((addr + len)>>8) & 0xff;
    // Pointer: Start of BASIC Variables
    if (mk2_writememory(buf, 2, 0x2d) < 0) {
        LOGERR("error writing to C64 memory.\n");
        exit(cleanup(EXIT_FAILURE));
    }
    // Pointer: Start of BASIC Arrays
    if (mk2_writememory(buf, 2, 0x2f) < 0) {
        LOGERR("error writing to C64 memory.\n");
        exit(cleanup(EXIT_FAILURE));
    }
    // Pointer: End of BASIC Arrays + 1
    if (mk2_writememory(buf, 2, 0x31) < 0) {
        LOGERR("error writing to C64 memory.\n");
        exit(cleanup(EXIT_FAILURE));
    }
    // Tape End Address/End of Program
    if (mk2_writememory(buf, 2, 0xae) < 0) {
        LOGERR("error writing to C64 memory.\n");
        exit(cleanup(EXIT_FAILURE));
    }
}

void execute_sys(int addr)
{
    unsigned char buf[10];
    int n;
    /* put SYS into keyboard buffer */
    buf[0] = 0;
    if (mk2_writememory(buf, 1, 198) < 0) {
        LOGERR("error writing to C64 memory.\n");
        exit(cleanup(EXIT_FAILURE));
    }
    buf[0] = 83;
    buf[1] = 217;
    buf[2] = 32;
    buf[3] = 32;
    buf[4] = 32;
    buf[5] = 32;
    buf[6] = 32;
    buf[7] = 32;
    addr &= 0xffff;
    n = sprintf((char*)&buf[2], "%d", addr);
    buf[n+2] = 13;
    if (mk2_writememory(buf, 8, 631) < 0) {
        LOGERR("error writing to C64 memory.\n");
        exit(cleanup(EXIT_FAILURE));
    }
    buf[0] = n+3;
    if (mk2_writememory(buf, 1, 198) < 0) {
        LOGERR("error writing to C64 memory.\n");
        exit(cleanup(EXIT_FAILURE));
    }
}

void execute_run(void)
{
    unsigned char buf[4];
    /* put RUN into keyboard buffer */
    buf[0] = 0;
    if (mk2_writememory(buf, 1, 198) < 0) {
        LOGERR("error writing to C64 memory.\n");
        exit(cleanup(EXIT_FAILURE));
    }
    buf[0] = 82;
    buf[1] = 85;
    buf[2] = 78;
    buf[3] = 13;
    if (mk2_writememory(buf, 4, 631) < 0) {
        LOGERR("error writing to C64 memory.\n");
        exit(cleanup(EXIT_FAILURE));
    }
    buf[0] = 4;
    if (mk2_writememory(buf, 1, 198) < 0) {
        LOGERR("error writing to C64 memory.\n");
        exit(cleanup(EXIT_FAILURE));
    }
}

void usage (void)
{
    printf(
    "mk2codenet %s (compiled on %s)\n"
    "usage: mk2codenet <options>\n"
    "\n"
    "-h --help              this help\n"
    "--verbose              enable verbose messages\n"
    "--debug                enable debug messages\n"
    "-p SerialPort          serial port to use. (default: "
#ifdef _WIN32
    "COM1)\n"
#else
    "ttyUSB0)\n"
#endif
    "\n"
    "-w Filename            Sends a PRG file to the C64.\n"
    "-wa Filename Addr      Sends a PRG file to the C64 to a specified address.\n"
    "-wb Filename Addr      Sends a binary file to the C64 to a specified address.\n"
    "-ra Filename Addr Len  Reads a PRG file from the C64 at a specified address.\n"
    "-rb Filename Addr Len  Reads a binary file from the C64 at a specified address.\n"
    "-x Filename            Sends a PRG file to the C64 and executes it\n"
    "-f Start End Fill      Fills a block of C64 memory.\n"
    "-e Addr                Jumps to an address in memory.\n"
    "-r                     Executes a program via \"RUN\"."
    "\n"
    "--reset                trigger target reset\n"
    "--resetwait            trigger target reset and wait until after reset\n"
    "\n"
    "-n, -T, -R             ignored for compatibility\n"
    , MK2CODENET_VERSION
    , __DATE__
    );
}

int cleanup(int n) 
{
    mk2_mainmenu();
    RS232_CloseComport(cport_nr);
    if (buffer) {
        free(buffer);
        buffer = NULL;
    }
    return n;
}

int main(int argc, char *argv[])
{
    char *name;
    int addr = 0, len = 0, end = 0, val = 0, i;
    FILE *f;
#ifdef _WIN32
    char *portname = "COM1";
#else
    char *portname = "ttyUSB0";
#endif

    if (argc < 2) {
        usage();
        exit (EXIT_FAILURE);
    }

    /* first check the options that should work before any other stuff is used */
    for (i = 1; i < argc; i++) {
        if (!strcmp("--verbose", argv[i])) {
            verbose = 1;
        } else if (!strcmp("--debug", argv[i])) {
            verbose = 2;
        } else if (!strcmp("-p", argv[i])) {
            i++;
            portname = argv[i];
        } else if (!strcmp("-h", argv[i]) || !strcmp("--help", argv[i]))  {
            usage();
            exit (EXIT_FAILURE);
        } else {
            break;
        }
    }

    if (i == argc) {
        usage();
        exit (EXIT_FAILURE);
    }

    cport_nr = RS232_GetPortnr(portname);

    if(RS232_OpenComport(cport_nr, MK2BAUDRATE, "8N1")) {
        LOGERR("error: Can not open comport\n");
        exit(EXIT_FAILURE);
    }

    if((buffer = (unsigned char*)malloc(C64_RAM_SIZE)) == NULL) {
        LOGERR("could not allocate memory.\n");
        exit(cleanup(EXIT_FAILURE));
    };

    /* check the rest of the options */
    for (; i < argc; i++) {
        if (!strcmp("-x", argv[i])) {
            /* write .prg file to memory + RUN */
            i++;
            name = argv[i];
            if ((f = fopen(name, "rb")) == NULL) {
                LOGERR("error opening: '%s'\n", name);
                exit(cleanup(EXIT_FAILURE));
            }
            addr = fgetc(f);
            addr += ((int)fgetc(f) << 8);

            len = fread(buffer, 1, C64_RAM_SIZE - addr, f);
            fclose(f);
            printf("sending '%s' ($%04x bytes to $%04x.)...\n", name, len, addr);
            if (mk2_writememory(buffer, len, addr) < 0) {
                LOGERR("error writing to C64 memory.\n");
                exit(cleanup(EXIT_FAILURE));
            }
            setup_pointers(addr, len);
            execute_run();
        } else if (!strcmp("-r", argv[i])) {
            /* execute prg via RUN */
            execute_run();
        } else if (!strcmp("-e", argv[i])) {
            /* execute prg via SYS */
            i++;
            addr = strtoul(argv[i], NULL, 0);
            execute_sys(addr);
        } else if (!strcmp("-w", argv[i])) {
            /* write .prg file to memory */
            i++;
            name = argv[i];
            if ((f = fopen(name, "rb")) == NULL) {
                LOGERR("error opening: '%s'\n", name);
                exit(cleanup(EXIT_FAILURE));
            }
            addr = fgetc(f);
            addr += ((int)fgetc(f) << 8);

            len = fread(buffer, 1, C64_RAM_SIZE - addr, f);
            fclose(f);
            printf("sending '%s' ($%04x bytes to $%04x.)...\n", name, len, addr);
            if (mk2_writememory(buffer, len, addr) < 0) {
                LOGERR("error writing to C64 memory.\n");
                exit(cleanup(EXIT_FAILURE));
            }
        } else if (!strcmp("-wa", argv[i])) {
            /* write .prg file to specified address */
            i++;
            name = argv[i];
            if ((f = fopen(name, "rb")) == NULL) {
                LOGERR("error opening: '%s'\n", name);
                exit(cleanup(EXIT_FAILURE));
            }
            i++;
            addr = strtoul(argv[i], NULL, 0);
            fgetc(f);fgetc(f);

            len = fread(buffer, 1, C64_RAM_SIZE - addr, f);
            fclose(f);
            printf("sending '%s' ($%04x bytes to $%04x.)...\n", name, len, addr);
            if (mk2_writememory(buffer, len, addr) < 0) {
                LOGERR("error writing to C64 memory.\n");
                exit(cleanup(EXIT_FAILURE));
            }
        } else if (!strcmp("-wb", argv[i])) {
            /* write binary file to specified address */
            i++;
            name = argv[i];
            if ((f = fopen(name, "rb")) == NULL) {
                LOGERR("error opening: '%s'\n", name);
                exit(cleanup(EXIT_FAILURE));
            }
            i++;
            addr = strtoul(argv[i], NULL, 0);

            len = fread(buffer, 1, C64_RAM_SIZE - addr, f);
            fclose(f);
            printf("sending '%s' ($%04x bytes to $%04x.)...\n", name, len, addr);
            if (mk2_writememory(buffer, len, addr) < 0) {
                LOGERR("error writing to C64 memory.\n");
                exit(cleanup(EXIT_FAILURE));
            }
        } else if (!strcmp("-ra", argv[i])) {
            /* read PRG file from specified address */
            i++;
            name = argv[i];
            if ((f = fopen(name, "wb")) == NULL) {
                LOGERR("error opening: '%s'\n", name);
                exit(cleanup(EXIT_FAILURE));
            }
            i++;
            addr = strtoul(argv[i], NULL, 0);
            i++;
            len = strtoul(argv[i], NULL, 0);

            printf("receiving '%s' ($%04x bytes from $%04x.)...\n", name, len, addr);
            if (mk2_readmemory(buffer, len, addr) < 0) {
                LOGERR("error reading from C64 memory.\n");
                exit(cleanup(EXIT_FAILURE));
            }

            fputc(addr & 0xff, f);
            fputc((addr >> 8) & 0xff, f);
            fwrite(buffer, 1, len, f);
            fclose(f);
            
        } else if (!strcmp("-rb", argv[i])) {
            /* read binary file from specified address */
            i++;
            name = argv[i];
            if ((f = fopen(name, "wb")) == NULL) {
                LOGERR("error opening: '%s'\n", name);
                exit(cleanup(EXIT_FAILURE));
            }
            i++;
            addr = strtoul(argv[i], NULL, 0);
            i++;
            len = strtoul(argv[i], NULL, 0);

            printf("receiving '%s' ($%04x bytes from $%04x.)...\n", name, len, addr);
            if (mk2_readmemory(buffer, len, addr) < 0) {
                LOGERR("error reading from C64 memory.\n");
                exit(cleanup(EXIT_FAILURE));
            }

            fwrite(buffer, 1, len, f);
            fclose(f);
            
        } else if (!strcmp("-f", argv[i])) {
            /* fill a block of memory */
            i++;
            addr = strtoul(argv[i], NULL, 0);
            i++;
            end = strtoul(argv[i], NULL, 0);
            i++;
            val = strtoul(argv[i], NULL, 0);
            len = (end - addr) + 1;

            memset(buffer, val, C64_RAM_SIZE);

            printf("filling $%04x to $%04x ($%04x bytes)...\n", addr, end, len);
            if (mk2_writememory(buffer, len, addr) < 0) {
                LOGERR("error writing to C64 memory.\n");
                exit(cleanup(EXIT_FAILURE));
            }
        } else if (!strcmp("-n", argv[i])) {
            /* ignored */
            i++;
        } else if (!strcmp("-T", argv[i])) {
            /* ignored */
            i++;
        } else if (!strcmp("-R", argv[i])) {
            /* ignored */
            i++;
        } else if (!strcmp("--reset", argv[i])) {
            execute_reset();
        } else if (!strcmp("--resetwait", argv[i])) {
            execute_resetwait();
        } else {
            usage();
            exit(cleanup(EXIT_FAILURE));
        }
    }
  
    return(cleanup(EXIT_SUCCESS));
}

